@@ -111,7 +111,6 @@ impl_Bonobo_ControlFrame_activateURI (PortableServer_Servant  servant,
 			 control_frame_signals [ACTIVATE_URI],
 			 (const char *) uri, (gboolean) relative);
 }
-
 static gint
 bonobo_control_frame_autoactivate_focus_in (GtkWidget     *widget,
 					    GdkEventFocus *focus,
@@ -144,7 +143,7 @@ bonobo_control_frame_autoactivate_focus_out (GtkWidget     *widget,
 
 
 static void
-bonobo_control_frame_socket_state_changed (GtkWidget    *socket,
+bonobo_control_frame_socket_state_changed (GtkWidget    *container,
 					   GtkStateType  previous_state,
 					   gpointer      user_data)
 {
@@ -155,7 +154,7 @@ bonobo_control_frame_socket_state_changed (GtkWidget    *socket,
 
 	bonobo_control_frame_control_set_state (
 		control_frame,
-		GTK_WIDGET_STATE (control_frame->priv->socket));
+		GTK_WIDGET_STATE (control_frame->priv->container));
 }
 
 
@@ -233,14 +232,6 @@ bonobo_control_frame_create_socket (BonoboControlFrame *control_frame)
 			    GTK_SIGNAL_FUNC (bonobo_control_frame_autoactivate_focus_out),
 			    control_frame);
 
-	/*
-	 * Setup a handler to proxy state changes.
-	 */
-	gtk_signal_connect (GTK_OBJECT (control_frame->priv->socket),
-			    "state_changed",
-			    bonobo_control_frame_socket_state_changed,
-			    control_frame);
-
 	/*
 	 * Setup a handler for socket destroy.
 	 */
@@ -304,6 +295,15 @@ bonobo_control_frame_construct (BonoboControlFrame *control_frame,
 	gtk_object_sink (GTK_OBJECT (control_frame->priv->container));
 	gtk_widget_show (control_frame->priv->container);
 
+	/*
+	 * Setup a handler to proxy state from the container.
+	 */
+	gtk_signal_connect (GTK_OBJECT (control_frame->priv->container),
+			    "state_changed",
+			    bonobo_control_frame_socket_state_changed,
+			    control_frame);
+
+
 	bonobo_control_frame_create_socket (control_frame);
 
 	return control_frame;
@@ -967,7 +967,7 @@ bonobo_control_frame_sync_unrealize (BonoboControlFrame *frame)
 }
 
 /**
- * bonobo_control_frame_focus:
+ * bonobo_control_frame_focus_child:
  * @frame: A control frame.
  * @direction: Direction in which to change focus.
  * 
@@ -979,7 +979,7 @@ bonobo_control_frame_sync_unrealize (BonoboControlFrame *frame)
  * passed on to the next widget.
  **/
 gboolean
-bonobo_control_frame_focus (BonoboControlFrame *frame, GtkDirectionType direction)
+bonobo_control_frame_focus_child (BonoboControlFrame *frame, GtkDirectionType direction)
 {
 	BonoboControlFramePrivate *priv;
 	CORBA_Environment ev;
@@ -1026,9 +1026,9 @@ bonobo_control_frame_focus (BonoboControlFrame *frame, GtkDirectionType directio
 
 	CORBA_exception_init (&ev);
 
-	result = Bonobo_Control_focus (priv->control, corba_direction, &ev);
+	result = Bonobo_Control_focusChild (priv->control, corba_direction, &ev);
 	if (BONOBO_EX (&ev)) {
-		g_message ("bonobo_control_frame_focus(): Exception while issuing focus "
+		g_message ("bonobo_control_frame_focus_child(): Exception while issuing focusChild "
 			   "request: `%s'", bonobo_exception_get_text (&ev));
 		result = FALSE;
 	}
